import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'dart:typed_data';

abstract class ConnectionInfo {
  static int IPSocket = 1;
  Connection _connection;
  int getType();
  Connection createConnection(ConnectionAction action);
  int send(int commandType, String text);
  bool initialized();
}
class IpSocketConnectionInfo extends ConnectionInfo {
  int localPort;
  String remoteAddress;
  int remotePort = -1;
  IpSocketConnectionInfo(this.localPort);
  void setRemote(String remoteAddress, int remotePort) {
    this.remoteAddress = remoteAddress;
    this.remotePort = remotePort;
  }

  int getType() { return ConnectionInfo.IPSocket; }
  Connection createConnection(ConnectionAction action) {
    _connection = new IPConnection(localPort);
    _connection.initialize(action);
    return _connection;
  }
  int send(int commandType, String text) {
    if (!initialized())
      throw new Exception('Not initialized');
    return _connection.send(commandType, text, remoteAddress, remotePort);
  }
  bool initialized() { return _connection != null && remoteAddress != null && remotePort > 0; }
}

abstract class ConnectionAction {
  Future onReceive(String localIp, int commandType, String command);
  Future onConnectionClosed(String localIp);
  Future onConnectionError(String localIp, String err);
  String jsModel2String(Map<String, dynamic> json) {
    return jsonEncode(json);
  }
  Map<String, dynamic> String2JsModel(String json) {
    return jsonDecode(json);
  }
}
abstract class Connection {
  String id;
  ConnectionAction _doSomething;
  Future initialize(ConnectionAction doSomething);
  int send(int commandType, String text, String clientAddress, int port);
}
class IPConnection extends Connection {
  int localPort;
  RawDatagramSocket _socket;
  IPConnection(this.localPort);
  static Future<String> getIp() async {
    for (var interface in await NetworkInterface.list()) {
      for (var addr in interface.addresses) {
        return addr.address;
      }
    }
    return null;
  }
  int _commandType = -1;
  List<int> _buffer = [];
  static int startCommand = 255;
  static int endCommand = 254;
  Future initialize(ConnectionAction doSomething) async {
    this._doSomething = doSomething;
    id = await getIp();
    try {
      Future.wait([RawDatagramSocket.bind(new InternetAddress(id), localPort)]).then((values) {
        _socket = values[0];
        _socket.listen((RawSocketEvent e) {
          try {
            switch (e) {
              case RawSocketEvent.read:
                Datagram dg = _socket.receive();
                for(int i= 0; i<dg.data.length; i++) {
                  if (dg.data[i] == startCommand) {
                    _buffer = [];
                    _commandType = -1;
                  } else if (dg.data[i] == endCommand) {
                    _doSomething.onReceive(id, _commandType, new Utf8Codec().decode(_buffer));
                    _buffer = [];
                    _commandType = -1;
                  } else if (_commandType == -1) {
                    _commandType = dg.data[i];
                  } else
                    _buffer.add(dg.data[i]);
                }
                _socket.writeEventsEnabled = true;
                break;
              case RawSocketEvent.write:
                break;
              case RawSocketEvent.closed:
                _doSomething.onConnectionClosed(id);
            }
          } catch (ex) {
            _doSomething.onConnectionError(id, ex);
          }
        });
      });
    } catch (ex) {
      _doSomething.onConnectionError(id, ex);
    }
  }
  int send(int commandType, String text, String clientAddress, int port) {
    List<int> command = [];
    command.add(startCommand);
    command.add(commandType);
    var c = new Utf8Codec().encode(text);
    for(var i=0; i<c.length; i++)
      command.add(c[i]);
    command.add(endCommand);
    return _socket.send(new Uint8List.fromList(command), new InternetAddress(clientAddress), port);
  }
}
